from typing import Optional

import numpy as np
from numpy.typing import NDArray


def _compute_auto_scales(
    functions: NDArray[np.float64], auto_scale: NDArray[np.bool_]
) -> Optional[NDArray[np.float64]]:
    if np.any(auto_scale):
        return np.where(auto_scale, np.fabs(functions[auto_scale]), 1.0)
    return None


def _get_failed_realizations(
    objectives: NDArray[np.float64],
    perturbed_objectives: Optional[NDArray[np.float64]],
    perturbation_min_success: int,
) -> NDArray[np.bool_]:
    failed_realizations = np.isnan(objectives[..., 0])
    if perturbed_objectives is not None:
        failed_pertubations = np.isnan(perturbed_objectives[..., 0])
        success_count = np.count_nonzero(~failed_pertubations, axis=-1)
        failed_realizations |= success_count < perturbation_min_success
    return np.array(failed_realizations)



"""Configuration class for objective functions."""

from __future__ import annotations

from typing import Optional

import numpy as np
from pydantic import model_validator

from ropt.config.utils import (
    Array1D,
    Array1DBool,
    Array1DInt,
    UniqueNames,
    broadcast_1d_array,
    broadcast_arrays,
    normalize,
)

from ._enopt_base_model import EnOptBaseModel


class ObjectiveFunctionsConfig(EnOptBaseModel):
    """The configuration class for objective functions.

    This configuration class defines objective functions configured by the
    `objective_functions` field in an
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    `ropt` supports optimization over multiple objectives, which are summed after
    weighting with values passed via the `weights` field. This field is a
    `numpy` array, with a length equal to the number of objective functions. Its
    values will be normalized to have a sum equal to 1. For example, when
    `weights` is set to `[1, 1]`, the stored values will be `[0.5, 0.5]`.

    The `names` field is optional. If given, the number of objective functions
    is set equal to its length. The `weights` array will then be broadcasted to
    the number of objective values. For example, if `names = ["f1", "f2"]` and
    `weights = 1.0`, the optimizer assumes two objective functions weighted by
    `[0.5, 0.5]`. If `names` is not set, the number of objectives is determined
    by the length of `weights`.

    The `scales` field contains scaling values for the objectives. These values
    are used to scale the objective function values to a desired order of
    magnitude. Each time new objective function values are obtained during
    optimization, they are divided by these values. The `auto_scale` field can
    be used to direct the optimizer to obtain an additional scaling by
    multiplying the values of `scales` by the values of the objective functions
    at the start of the optimization. Both the `scales` and `auto_scale` arrays
    will be broadcasted to have a size equal to the number of objective
    functions.

    Info:
        Both the `scales` values and the values obtained by auto-scaling will be
        applied. Thus, if `scales` is not supplied, auto-scaling will scale the
        objectives such that their initial values will be equal to one. Setting
        `scales` also allows for scaling to different initial values.

    The objective functions may be subject to realization filters and function
    transforms. The `realization_filters` and `function_transforms` fields
    contain indices to the realization filter or function transform objects to
    use. The objects referred to are configured in the parent
    [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.

    Attributes:
        names:               Optional names of the objective functions
        weights:             Objective functions weights (default: 1.0)
        scales:              The scaling factors (default: 1.0)
        auto_scale:          Enable/disable auto-scaling (default: `False`)
        realization_filters: Optional realization filter indices
        function_transforms: Optional function transform indices
    """

    names: Optional[UniqueNames] = None
    weights: Array1D = np.array(1.0)
    scales: Array1D = np.array(1.0, dtype=np.float64)
    auto_scale: Array1DBool = np.array(False)  # noqa: FBT003
    realization_filters: Optional[Array1DInt] = None
    function_transforms: Optional[Array1DInt] = None

    @model_validator(mode="after")
    def _broadcast_and_normalize(self) -> ObjectiveFunctionsConfig:
        if self.names is not None:
            size = len(self.names)
            for name in ("scales", "auto_scale", "weights"):
                setattr(self, name, broadcast_1d_array(getattr(self, name), name, size))
        else:
            self.scales, self.auto_scale, self.weights = broadcast_arrays(
                self.scales, self.auto_scale, self.weights
            )

        self.weights = normalize(self.weights)
        return self



import time
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from itertools import count
from typing import Any, DefaultDict, Dict, List, Optional, Tuple

import numpy as np
from numpy.typing import NDArray

from ._evaluator import EvaluatorContext, EvaluatorResult


@dataclass
class ConcurrentTask(ABC):
    """Abstract data class for tasks in a concurrent evaluator.

    These task objects should contain the future that represents the task and
    return the results of the evaluation.

    It should implement two methods to retrieve the objective function values
    and optionally the constraint values.
    """

    future: Optional[Any]

    @abstractmethod
    def get_objectives(self) -> Optional[NDArray[np.float64]]:
        """Return the objectives calculated by the task.

        This method will only be called after the future is done, and if no
        exception was raised during the execution of the task.

        Info:
            The return value might be None, as derived evaluator classes might
            also employ tasks that do not return results.

        Returns:
            The calculated objectives.
        """

    def get_constraints(self) -> Optional[NDArray[np.float64]]:
        """Return the constraints calculated by the task, if available.

        This method will only be called after the future is done, and if no
        exception was raised during the execution of the task.

        This has a default implementation that returns `None`, which is only
        usable when it is certain that there are no non-linear constraints. If
        there are non-linear constraints, this method should be overridden.

        Returns:
            The calculated constraints or `None`.
        """
        return None


class ConcurrentEvaluator(ABC):
    """Abstract base class for implementing a concurrent evaluator.

    This abstract base class provides the framework for implementing an
    evaluator that uses a concurrent executor.

    The `launch` method must be implemented to start an evaluation for one
    vector of variables, and return a future-like object, compatible with the
    `concurrent.futures` module of Python, implementing at least the `done()`,
    `exception()`, and `result()` methods. The `monitor` method can be overloaded
    to implement specific monitoring functionality, the default implementation does
    nothing.

    The class implements an optional caching mechanism to prevent repeated
    evaluation of functions. When enabled, all evaluation results are cached in
    memory and reused when requested. This is in particular useful in
    optimization plans with multiple or nested optimizations where often
    restarts occur from points that already have been evaluated before.
    """

    def __init__(
        self, *, enable_cache: bool = True, polling: float = 0.1, max_submit: int = 500
    ) -> None:
        """Initialize a concurrent evaluator object.

        Some general properties of the evaluator are fixed at initialization time:

        - The cache may be enabled or disabled using the `enable_cache` parameter.
        - While evaluations are running concurrently the evaluator will
          regularly poll them to check their status. The `polling` parameter
          determine the delay between polling events.
        - When a very large number of jobs is submitted this may overwhelm the
          evaluator, depending on the implementation. To prevent this, the jobs
          may be submitted in smaller portions, the size of which are defined by
          the `max_submit` argument.

        Args:
            enable_cache: Enable the caching mechanism
            polling:      Time in seconds between checking job status
            max_submit:   Maximum number of variables to submit simultaneously
        """
        self._batch_id = 0
        self._polling = polling
        self._max_submit = max_submit
        self._cache: Optional[_Cache] = _Cache() if enable_cache else None

    @abstractmethod
    def launch(
        self,
        batch_id: Any,  # noqa: ANN401
        job_id: int,
        variables: NDArray[np.float64],
        context: EvaluatorContext,
    ) -> Optional[ConcurrentTask]:
        """Launch an evaluation return a future for each job.

        This method must implement the process of launching a a single function
        evaluation for a variable vector passed via the `variables` parameter. A
        unique batch ID is passed via the `batch_id`, which can be optionally
        used by the evaluator to identify the current batch of functions.

        This method should return a dictionary mapping the indices of the jobs
        to the tasks that will contain the result. The tasks are objects
        deriving from the [`ConcurrentTask`][ropt.evaluator.ConcurrentTask]
        class, containing the future object representing the launched task and
        its result. Under the hood, other tasks may be launched, but only those
        that contain results should be returned.

        This method is called by the `__call__` method, which implements the
        [`Evaluator`][ropt.evaluator.Evaluator] callback signature.

        The `context` argument with optional information is passed from the
        `__call__` method unchanged.

        Args:
            batch_id:  The ID of the batch of evaluations to run.
            job_id:    The ID of the job launched for the variables
            variables: The matrix of variables to evaluate.
            context:   Evaluator context.

        Returns:
            The last future or `None` if no tasks were launched.
        """

    def monitor(self) -> None:  # noqa: B027
        """Monitor the states of the running evaluations.

        This method is called regularly in the polling loop that checks the
        futures until all results are collected. If the status of the
        evaluations should be monitored, this method should be overridden.

        The time in seconds between calls in the polling loop can be modified by
        setting the `polling` attribute upon object initialization.
        """

    def __call__(
        self, variables: NDArray[np.float64], context: EvaluatorContext
    ) -> EvaluatorResult:
        """Launch all evaluations and collect the results.

        This method implements the evaluation function, passed to the ensemble
        optimizer constructor.

        Args:
            variables: The matrix of variables to evaluate.
            context:   Evaluator context

        Returns:
            The batch ID and objective and constraint function values.
        """
        objective_results, constraint_results = _init_results(variables, context)

        active = self._try_cache(
            variables, context, objective_results, constraint_results
        )

        var_idx = 0
        tasks: Dict[int, ConcurrentTask] = {}

        # Keep submitting and monitoring until all variables and tasks are done:
        while var_idx < variables.shape[0] or len(tasks) > 0:
            # Add more tasks up to a maximum of max_submit:
            for _ in range(self._max_submit - len(tasks)):
                if var_idx < variables.shape[0]:
                    if active is None or active[var_idx]:
                        task = self.launch(
                            self._batch_id, var_idx, variables[var_idx], context
                        )
                        if task is not None:
                            tasks[var_idx] = task
                    var_idx += 1

            # Monitor the current tasks:
            for idx in list(tasks.keys()):  # tasks changes size, hence list()
                future = tasks[idx].future
                if future is None or future.done():
                    if future is None or future.exception() is None:
                        objective_results[idx, :] = tasks[idx].get_objectives()
                        if constraint_results is not None:
                            constraint_results[idx, :] = tasks[idx].get_constraints()
                    del tasks[idx]
            self.monitor()
            time.sleep(self._polling)

        result = EvaluatorResult(
            objectives=objective_results,
            constraints=constraint_results,
            batch_id=self._batch_id,
        )

        self._update_cache(
            variables, context, active, objective_results, constraint_results
        )

        self._batch_id += 1
        return result

    def _try_cache(
        self,
        variables: NDArray[np.float64],
        context: EvaluatorContext,
        objectives: NDArray[np.float64],
        constraints: Optional[NDArray[np.float64]],
    ) -> Optional[NDArray[np.bool_]]:
        if self._cache is None:
            return context.active
        active = (
            np.ones(variables.shape[0], dtype=np.bool_)
            if context.active is None
            else np.fromiter(
                (context.active[realization] for realization in context.realizations),
                dtype=np.bool_,
            )
        )

        for job_idx, real_id in enumerate(self._get_realization_ids(context)):
            cache_id = self._cache.find_key(real_id, variables[job_idx, ...])
            if cache_id is not None:
                active[job_idx] = False
                objectives[job_idx, ...] = self._cache.get_objectives(cache_id)
                if constraints is not None:
                    constraints[job_idx, ...] = self._cache.get_constraints(cache_id)

        return active

    def _update_cache(
        self,
        variables: NDArray[np.float64],
        context: EvaluatorContext,
        active: Optional[NDArray[np.bool_]],
        objectives: NDArray[np.float64],
        constraints: Optional[NDArray[np.float64]],
    ) -> None:
        if self._cache is not None:
            assert active is not None
            for job_idx, real_id in enumerate(self._get_realization_ids(context)):
                if active[job_idx]:
                    self._cache.add_simulation_results(
                        job_idx, real_id, variables, objectives, constraints
                    )

    def _get_realization_ids(self, context: EvaluatorContext) -> Tuple[Any, ...]:
        names = context.config.realizations.names
        if names is None:
            names = tuple(range(context.config.realizations.weights.size))
        return tuple(names[realization] for realization in context.realizations)


def _init_results(
    variables: NDArray[np.float64], context: EvaluatorContext
) -> Tuple[NDArray[np.float64], Optional[NDArray[np.float64]]]:
    objective_count = context.config.objective_functions.weights.size
    constraint_count = (
        0
        if context.config.nonlinear_constraints is None
        else context.config.nonlinear_constraints.rhs_values.size
    )
    objective_results = np.full(
        (variables.shape[0], objective_count), fill_value=np.nan, dtype=np.float64
    )
    if constraint_count > 0:
        constraint_results = np.full(
            (variables.shape[0], constraint_count), fill_value=np.nan, dtype=np.float64
        )
    else:
        constraint_results = None
    if context.active is not None:
        inactive = ~context.active
        objective_results[inactive, :] = 0.0
        if constraint_results is not None:
            constraint_results[inactive, :] = 0.0
    return objective_results, constraint_results


class _Cache:
    def __init__(self) -> None:
        # Stores the realization/controls key, together with an ID.
        self._keys: DefaultDict[int, List[Tuple[NDArray[np.float64], int]]] = (
            defaultdict(list)
        )
        # Store objectives and constraints by ID:
        self._objectives: Dict[int, NDArray[np.float64]] = {}
        self._constraints: Dict[int, NDArray[np.float64]] = {}

        # Generate unique ID's:
        self._counter = count()

    def add_simulation_results(
        self,
        job_idx: int,
        real_id: int,
        control_values: NDArray[np.float64],
        objectives: NDArray[np.float64],
        constraints: Optional[NDArray[np.float64]],
    ) -> None:
        cache_id = next(self._counter)
        self._keys[real_id].append((control_values[job_idx, :].copy(), cache_id))
        self._objectives[cache_id] = objectives[job_idx, ...].copy()
        if constraints is not None:
            self._constraints[cache_id] = constraints[job_idx, ...].copy()

    def find_key(
        self, real_id: int, control_vector: NDArray[np.float64]
    ) -> Optional[int]:
        # Brute-force search, premature optimization is the root of all evil:
        for cached_vector, cache_id in self._keys.get(real_id, []):
            if np.allclose(
                control_vector,
                cached_vector,
                rtol=0.0,
                atol=float(np.finfo(np.float32).eps),
            ):
                return cache_id
        return None

    def get_objectives(self, cache_id: int) -> NDArray[np.float64]:
        return self._objectives[cache_id]

    def get_constraints(self, cache_id: int) -> NDArray[np.float64]:
        return self._constraints[cache_id]



"""Utilities for checking configuration values."""

import sys
from collections import Counter
from enum import IntEnum
from typing import Any, Optional, Tuple, Type, cast

import numpy as np
from numpy.typing import ArrayLike, NDArray
from pydantic import BeforeValidator

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


def normalize(array: NDArray[np.float64]) -> NDArray[np.float64]:
    """Normalize a vector.

        Normalize the sum of the values to one.

    Args:
        array: The input array

    Returns:
        ValueError: The normalized array
    """
    if array.sum() < np.finfo(np.float64).eps:
        msg = "the sum of weights is not positive"
        raise ValueError(msg)
    return immutable_array(array / array.sum())


def immutable_array(
    array_like: ArrayLike,
    **kwargs: Any,  # noqa: ANN401
) -> NDArray[Any]:
    """Make an immutable array.

    Converts the input to an array and makes it immutable.`

    Args:
        array_like: The input
        kwargs    : Additional keyword arguments for array conversion

    Returns:
        The immutable array.
    """
    array = np.array(array_like, **kwargs)
    array.setflags(write=False)
    return array


def broadcast_arrays(*args: Any) -> Tuple[NDArray[Any], ...]:  # noqa: ANN401
    """Broadcast a set of arrays and makes them immutable.

    Args:
        args: The input arrays

    Returns:
        The broadcasted immutable arrays.
    """
    results = np.broadcast_arrays(*args)
    return tuple(immutable_array(result) for result in results)


def broadcast_1d_array(array: NDArray[Any], name: str, size: int) -> NDArray[Any]:
    """Broadcast the input array to an 1D array of given size.

    Args:
        array: The input array
        name:  The name of the array, used in an error message
        size:  The size of the result

    Returns:
        An 1D array of the requested size.
    """
    if size == 0:
        return immutable_array([], dtype=array.dtype)
    try:
        return np.broadcast_to(immutable_array(array), (size,))
    except ValueError as err:
        msg = f"{name} cannot be broadcasted to a length of {size}"
        raise ValueError(msg) from err


def check_enum_values(value: NDArray[np.ubyte], enum_type: Type[IntEnum]) -> None:
    """Check if an enum value is valid.

    Given an array of byte integers, check of the values are within the range of
    values of the given enum.

    Args:
        value:     The enum values
        enum_type: The type to check

    Raises:
        ValueError: If the array contains an invalid value.
    """
    min_enum = min(item.value for item in enum_type)
    max_enum = max(item.value for item in enum_type)
    if np.any(value < min_enum) or np.any(value > max_enum):
        msg = "invalid enumeration value"
        raise ValueError(msg)


def _convert_1d_array(array: Optional[ArrayLike]) -> Optional[NDArray[np.float64]]:
    if array is None:
        return array
    return immutable_array(array, dtype=np.float64, ndmin=1)


def _convert_1d_array_intc(array: Optional[ArrayLike]) -> Optional[NDArray[np.intc]]:
    if array is None:
        return array
    return immutable_array(array, dtype=np.intc, ndmin=1)


def _convert_1d_array_bool(
    array: Optional["ArrayLike"],
) -> Optional[NDArray[np.bool_]]:
    if array is None:
        return array
    return immutable_array(array, dtype=np.bool_, ndmin=1)


def _convert_2d_array(array: Optional[ArrayLike]) -> Optional[NDArray[np.float64]]:
    if array is None:
        return array
    return immutable_array(array, dtype=np.float64, ndmin=2)


def _convert_enum_array(array: Optional[ArrayLike]) -> Optional[NDArray[np.ubyte]]:
    if array is None:
        return array
    return immutable_array(array, dtype=np.ubyte, ndmin=1)


def _convert_indices(array: Optional[ArrayLike]) -> Optional[NDArray[np.intc]]:
    if array is None:
        return array
    return cast(
        NDArray[np.intc],
        np.unique(immutable_array(array, dtype=np.intc, ndmin=1)),
    )


def _check_duplicates(names: Optional[Tuple[Any, ...]]) -> Optional[Tuple[Any, ...]]:
    if names is not None:
        counts = Counter(names)
        duplicates = [name for name, count in counts.items() if count > 1]
        if duplicates:
            raise ValueError("duplicate names: " + ", ".join(duplicates))
    return names


if sys.version_info >= (3, 9):
    Array1D = Annotated[NDArray[np.float64], BeforeValidator(_convert_1d_array)]
    Array2D = Annotated[NDArray[np.float64], BeforeValidator(_convert_2d_array)]
    ArrayIndices = Annotated[NDArray[np.intc], BeforeValidator(_convert_indices)]
    ArrayEnum = Annotated[NDArray[np.ubyte], BeforeValidator(_convert_enum_array)]
    Array1DInt = Annotated[NDArray[np.intc], BeforeValidator(_convert_1d_array_intc)]
    Array1DBool = Annotated[NDArray[np.bool_], BeforeValidator(_convert_1d_array_bool)]
    UniqueNames = Annotated[Tuple[Any, ...], BeforeValidator(_check_duplicates)]
else:
    Array1D = Annotated[ArrayLike, BeforeValidator(_convert_1d_array)]
    Array2D = Annotated[ArrayLike, BeforeValidator(_convert_2d_array)]
    ArrayIndices = Annotated[ArrayLike, BeforeValidator(_convert_indices)]
    ArrayEnum = Annotated[ArrayLike, BeforeValidator(_convert_enum_array)]
    Array1DInt = Annotated[ArrayLike, BeforeValidator(_convert_1d_array_intc)]
    Array1DBool = Annotated[ArrayLike, BeforeValidator(_convert_1d_array_bool)]
    UniqueNames = Annotated[Tuple[Any, ...], BeforeValidator(_check_duplicates)]



"""Classes for reporting optimization results.

The classes in this module are designed to gather multiple results generated
during the execution of an optimization plan. Currently, the results can
be reported as a [`pandas`](https://pandas.pydata.org/) DataFrame using the
[`ResultsDataFrame`][ropt.report.ResultsDataFrame] class, or as a text file in
a tabular format using the [`ResultsTable`][ropt.report.ResultsTable] class.
"""

from ._data_frame import ResultsDataFrame
from ._table import ResultsTable

__all__ = [
    "ResultsDataFrame",
    "ResultsTable",
]



"""Configuration classes for optimization plans.

Optimization plans are defined an run by [`Plan`][ropt.plan.Plan] objects.
"""

from ._plan_config import ContextConfig, PlanConfig, StepConfig

__all__ = [
    "ContextConfig",
    "StepConfig",
    "PlanConfig",
]



from typing import Any, Dict

import numpy as np
import pytest

from ropt.config.enopt import EnOptConfig
from ropt.enums import ResultAxisName
from ropt.results import (
    FunctionEvaluations,
    FunctionResults,
    Functions,
    GradientEvaluations,
    GradientResults,
    Gradients,
    Realizations,
)

pandas = pytest.importorskip("pandas")
from ropt.results._pandas import _to_series


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "names": ["x", "y"],
            "initial_values": [0.0, 0.0],
        },
        "objective_functions": {
            "names": ["f1", "f2"],
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "names": ["r1", "r2", "r3"],
            "weights": [1.0] * 3,
        },
        "gradient": {
            "number_of_perturbations": 5,
        },
    }


@pytest.fixture(name="function_result")
def function_result_fixture(enopt_config: Any) -> FunctionResults:
    config = EnOptConfig.model_validate(enopt_config)
    evaluations = FunctionEvaluations.create(
        config=config,
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    realizations = Realizations(
        objective_weights=np.arange(6, dtype=np.float64).reshape((2, 3)),
        failed_realizations=np.zeros(3, dtype=np.bool_),
    )
    functions = Functions.create(
        config=config, weighted_objective=np.array(1.0), objectives=np.array([1.0, 2.0])
    )
    return FunctionResults(
        result_id=0,
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=realizations,
        functions=functions,
    )


@pytest.fixture(name="gradient_result")
def gradient_result_fixture() -> GradientResults:
    evaluations = GradientEvaluations(
        variables=np.array([1.0, 2.0]),
        perturbed_variables=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
        perturbed_objectives=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
    )
    gradients = Gradients(
        weighted_objective=np.array([1.0, 2.0]),
        objectives=np.arange(4, dtype=np.float64).reshape((2, 2)),
    )
    return GradientResults(
        result_id=0,
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=Realizations(
            failed_realizations=np.zeros(36, dtype=np.bool_),
        ),
        gradients=gradients,
    )


def test__to_series(enopt_config: Any, gradient_result: GradientResults) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    series = _to_series(
        config, gradient_result.evaluations, 0, 1, "perturbed_variables"
    )
    assert series is not None
    assert len(series) == gradient_result.evaluations.perturbed_variables.size
    assert series.index.names == [
        "result_id",
        "batch_id",
        "realization",
        "perturbation",
        "variable",
    ]
    assert config.variables.names is not None
    assert config.realizations.names is not None
    for v_idx, var in enumerate(config.variables.names):
        for r_idx, real in enumerate(config.realizations.names):
            for pert in range(config.gradient.number_of_perturbations):
                assert (
                    series.loc[(0, 1, real, pert, var)]
                    == gradient_result.evaluations.perturbed_variables[
                        r_idx,
                        pert,
                        v_idx,
                    ]
                )


def test_to_dataframe(enopt_config: Any, gradient_result: GradientResults) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    frame = gradient_result.to_dataframe(config, "evaluations")
    assert len(frame) == gradient_result.evaluations.perturbed_variables.size * 2
    assert set(frame.index.names) == {
        "result_id",
        "batch_id",
        "realization",
        "perturbation",
        "variable",
        "objective",
    }


def test_to_dataframe_formatter(
    enopt_config: Any, function_result: FunctionResults
) -> None:
    enopt_config["variables"]["names"] = [("x", 0), ("x", 1)]
    config = EnOptConfig.model_validate(enopt_config)
    frame = function_result.to_dataframe(config, "evaluations", select=["variables"])
    assert (
        tuple(frame.index.get_level_values("variable"))
        == config.variables.get_formatted_names()
    )
    assert list(frame.index.get_level_values("variable")) == ["x:0", "x:1"]


def test_to_dataframe_unstack(
    enopt_config: Any, gradient_result: GradientResults
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    frame = gradient_result.to_dataframe(
        config,
        "evaluations",
        select=["perturbed_variables"],
        unstack=[ResultAxisName.REALIZATION, ResultAxisName.VARIABLE],
    )
    assert frame.index.names == ["result_id", "batch_id", "perturbation"]
    assert list(frame.columns.values) == [
        ("perturbed_variables", "r1", "x"),
        ("perturbed_variables", "r1", "y"),
        ("perturbed_variables", "r2", "x"),
        ("perturbed_variables", "r2", "y"),
        ("perturbed_variables", "r3", "x"),
        ("perturbed_variables", "r3", "y"),
    ]


def test_to_dataframe_unstack2(
    enopt_config: Any, gradient_result: GradientResults
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    assert gradient_result.gradients is not None
    frame = gradient_result.to_dataframe(
        config,
        "gradients",
        select=["objectives", "weighted_objective"],
        unstack=[ResultAxisName.OBJECTIVE, ResultAxisName.VARIABLE],
    )
    assert list(frame.columns.values) == [
        ("objectives", "f1", "x"),
        ("objectives", "f1", "y"),
        ("objectives", "f2", "x"),
        ("objectives", "f2", "y"),
        ("weighted_objective", "x"),
        ("weighted_objective", "y"),
    ]


def test_to_dataframe_unstack_only_variable(
    enopt_config: Any, gradient_result: GradientResults
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    frame = gradient_result.to_dataframe(
        config,
        "evaluations",
        select=["perturbed_objectives", "perturbed_variables"],
        unstack=[ResultAxisName.VARIABLE],
    )
    assert frame.index.names == [
        "result_id",
        "batch_id",
        "realization",
        "perturbation",
        "objective",
    ]
    assert list(frame.columns.values) == [
        "perturbed_objectives",
        ("perturbed_variables", "x"),
        ("perturbed_variables", "y"),
    ]


def test_to_dataframe_join(enopt_config: Any, function_result: FunctionResults) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    frame1 = function_result.to_dataframe(config, "evaluations")
    frame2 = function_result.to_dataframe(config, "functions")
    frame1.columns = pandas.Index(
        "_".join(column) if isinstance(column, tuple) else column
        for column in frame1.columns.to_numpy()
    )
    frame2.columns = pandas.Index(
        "_".join(column) if isinstance(column, tuple) else column
        for column in frame2.columns.to_numpy()
    )
    frame = frame1.join(frame2, how="inner", lsuffix="_eval", rsuffix="_func")
    assert not frame.empty
    assert len(frame.columns) == len(frame1.columns) + len(frame2.columns)



"""This file defines the event objects that steps may use."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

if TYPE_CHECKING:
    from ropt.results import Results


@dataclass
class ContextUpdate:
    """Base class for data objects used for updating context objects.

    This class is used as a base for classes that store data produced by
    optimization plan steps und used to update context objects.

    Attributes:
        step_name: Optional name of the step generating the data object
    """

    step_name: Optional[str]


@dataclass
class ContextUpdateDict(ContextUpdate):
    """Context update data class used to represent dictionary data.

    Attributes:
        data: Generic data represented by a dictionary.
    """

    data: Dict[str, Any]


@dataclass
class ContextUpdateResults(ContextUpdate):
    """Context update data class used for updating with new results.

    Attributes:
        results: The new results
    """

    results: Tuple[Results, ...] = field(default_factory=tuple)



"""The plugin manager."""

from __future__ import annotations

import sys
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Dict, Final, Literal

from ropt.exceptions import ConfigError

from .function_transform.base import FunctionTransformPlugin
from .optimizer.base import OptimizerPlugin
from .plan.base import PlanPlugin
from .realization_filter.base import RealizationFilterPlugin
from .sampler.base import SamplerPlugin

if TYPE_CHECKING:
    from ropt.plugins.base import Plugin

if sys.version_info >= (3, 10):
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points


_PLUGIN_TYPES: Final = {
    "function_transform": FunctionTransformPlugin,
    "optimizer": OptimizerPlugin,
    "sampler": SamplerPlugin,
    "realization_filter": RealizationFilterPlugin,
    "plan": PlanPlugin,
}

PluginType = Literal[
    "optimizer",
    "sampler",
    "realization_filter",
    "function_transform",
    "plan",
]
""" Plugin Types Supported by `ropt`"""


class PluginManager:
    """The plugin manager."""

    def __init__(self) -> None:
        """Initialize the plugin manager.

        The plugin manager object is initialized via the entry points mechanism
        (see
        [`importlib.metadata`](https://docs.python.org/3/library/importlib.metadata.html)).

        For instance, to install an additional optimizer plugin, implemented in
        an independent package, and assuming installation via a `pyproject.toml`
        file, add the following:

        ```toml
        [project.entry-points."ropt.plugins.optimizer"]
        my_optimizer = "my_optimizer_pkg.my_plugin:MyOptimizer"
        ```
        This will make the `MyOptimizer` class from the `my_optimizer_pkg`
        package available under the name `my_optimizer`. The `MyOptimizer` class
        will be used to create
        [`Optimizer`][ropt.plugins.optimizer.base.Optimizer] objects and to
        faciliate this, it should derive from the
        [`OptimizerPlugin`][ropt.plugins.optimizer.base.OptimizerPlugin] class.

        Plugins can also be added dynamically using the `add_plugins` method.
        """
        # Built-in plugins, listed for all possible plugin types:
        self._plugins: Dict[PluginType, Dict[str, Plugin]] = {
            "optimizer": {},
            "sampler": {},
            "realization_filter": {},
            "function_transform": {},
            "plan": {},
        }

        for plugin_type in self._plugins:
            self.add_plugins(plugin_type, _from_entry_points(plugin_type))

    def add_plugins(self, plugin_type: PluginType, plugins: Dict[str, Plugin]) -> None:
        """Add a plugin at runtime.

        This method adds one or more plugins of a specific type to the plugin
        manager. The `plugins` argument maps the names of the new plugins to
        [`Plugin`][ropt.plugins.Plugin] object.

        The plugin names are case-insensitive.

        Note: Plugin types


        Args:
            plugin_type: Type of the plugin.
            plugins:     Dictionary of plugins.
        """
        for name, plugin in plugins.items():
            name_lower = name.lower()
            if name_lower in self._plugins[plugin_type]:
                msg = f"Duplicate plugin name: {name_lower}"
                raise ConfigError(msg)
            self._plugins[plugin_type][name_lower] = plugin

    def get_plugin(self, plugin_type: PluginType, method: str) -> Any:  # noqa: ANN401
        """Retrieve a plugin by type and method name.

        Args:
            plugin_type: The type of the plugin to retrieve.
            method:      The name of the method the plugin should provide.
        """
        plugin_name, sep, method_name = method.rpartition("/")

        if sep == "/":
            assert method_name != ""
            plugin = self._plugins[plugin_type].get(plugin_name.lower())
            if plugin is None:
                msg = f"Plugin not found: {method}"
                raise ConfigError(msg)
            return plugin

        for plugin in self._plugins[plugin_type].values():
            if plugin.is_supported(method_name):
                return plugin

        msg = f"Method not found: {method}"
        raise ConfigError(msg)

    def is_supported(self, plugin_type: PluginType, method: str) -> bool:
        """Check if a method is supported.

        Args:
            plugin_type: The type of the plugin to retrieve.
            method:      The name of the method the plugin should provide.
        """
        _, _, method_name = method.rpartition("/")
        try:
            plugin = self.get_plugin(plugin_type, method)
        except ConfigError:
            return False
        return bool(plugin.is_supported(method_name))


@lru_cache  # Without the cache, repeated calls are very slow
def _from_entry_points(plugin_type: str) -> Dict[str, Plugin]:
    plugins: Dict[str, Plugin] = {}
    for entry_point in entry_points().select(group=f"ropt.plugins.{plugin_type}"):
        plugin = entry_point.load()
        plugins[entry_point.name] = plugin()
        if not isinstance(plugins[entry_point.name], _PLUGIN_TYPES[plugin_type]):
            msg = (
                f"Incorrect type for {plugin_type} plugin `{entry_point.name}`"
                f": {type(plugins[entry_point.name])}"
            )
            raise TypeError(msg)
    return plugins



from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional

import pytest

from ropt.config.enopt import EnOptConfig
from ropt.enums import EventType
from ropt.events import Event
from ropt.plan import OptimizationPlanRunner
from ropt.report import ResultsTable

# Requires pandas:
pd = pytest.importorskip("pandas")


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "speculative": True,
            "max_functions": 3,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
            "upper_bounds": 1.0,
            "lower_bounds": -1.0,
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def _handle_results(event: Event, reporter: ResultsTable, config: EnOptConfig) -> None:
    assert event.results is not None
    reporter.add_results(config, event.results)


def test_tabular_report_no_results(
    enopt_config: Any, evaluator: Any, tmp_path: Path
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsTable({}, path=tmp_path / "results.txt")
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()
    assert not Path(tmp_path / "results.txt").exists()


def test_tabular_report_results(
    enopt_config: Any, evaluator: Any, tmp_path: Path
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsTable(
        {
            "result_id": "eval-ID",
            "evaluations.variables": "Variables",
        },
        path=tmp_path / "results.txt",
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()

    assert Path(tmp_path / "results.txt").exists()
    results = pd.read_fwf(tmp_path / "results.txt", header=[0, 1], skiprows=[2])

    assert (
        results.columns.get_level_values(level=0).to_list()
        == ["eval-ID"] + ["Variables"] * 3
    )
    assert results.columns.get_level_values(level=1)[1:].to_list() == [
        "0",
        "1",
        "2",
    ]
    assert len(results) == 3


def test_tabular_report_data_frames_results_formatted_names(
    enopt_config: Any, evaluator: Any, tmp_path: Path
) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsTable(
        {
            "result_id": "eval-ID",
            "evaluations.variables": "Variables",
        },
        path=tmp_path / "results.txt",
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()

    assert Path(tmp_path / "results.txt").exists()
    results = pd.read_fwf(tmp_path / "results.txt", header=[0, 1], skiprows=[2])
    assert results.columns.get_level_values(level=0)[1:].to_list() == ["Variables"] * 3
    assert list(results.columns.get_level_values(level=1)[1:].to_list()) == [
        f"a:{idx + 1}" for idx in range(3)
    ]


def test_tabular_report_data_frames_gradients(
    enopt_config: Any, evaluator: Any, tmp_path: Path
) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsTable(
        {
            "result_id": "eval-ID",
            "gradients.weighted_objective": "Total Objective",
        },
        tmp_path / "gradients.txt",
        table_type="gradients",
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()
    assert Path(tmp_path / "gradients.txt").exists()
    gradients = pd.read_fwf(tmp_path / "gradients.txt", header=[0, 1], skiprows=[2])
    assert (
        gradients.columns.get_level_values(level=0).to_list()
        == ["eval-ID"] + ["Total Objective"] * 3
    )
    assert gradients.columns.get_level_values(level=1)[1:].to_list() == [
        "a:1",
        "a:2",
        "a:3",
    ]
    assert len(gradients) == 3


@pytest.mark.parametrize("min_header_len", [None, 4])
def test_tabular_report_data_frames_min_header_len(
    enopt_config: Any, evaluator: Any, tmp_path: Path, min_header_len: Optional[int]
) -> None:
    enopt_config["variables"]["names"] = [("a", 1), ("a", 2), ("a", 3)]
    config = EnOptConfig.model_validate(enopt_config)
    reporter = ResultsTable(
        {
            "result_id": "eval-ID",
            "evaluations.variables": "Variables",
        },
        path=tmp_path / "results.txt",
        min_header_len=min_header_len,
    )
    OptimizationPlanRunner(config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION,
        partial(_handle_results, reporter=reporter, config=config),
    ).run()

    assert Path(tmp_path / "results.txt").exists()
    with Path.open(tmp_path / "results.txt") as fp:
        lines = fp.readlines()
    header_idx = 2 if min_header_len is None else min_header_len
    assert lines[header_idx][0] == "-"
    results = pd.read_fwf(
        tmp_path / "results.txt",
        header=list(range(header_idx)),
        skip_blank_lines=False,
        skiprows=[header_idx],
    )
    assert len(results.columns[0]) == header_idx
    assert len(results) == 3



"""Differential evaluation optimization example.

This example uses the differential evolution method to solve a discrete
problem with a linear constraint.
"""

from typing import Any, Dict

import numpy as np
from numpy.typing import NDArray

from ropt.enums import ConstraintType, EventType, VariableType
from ropt.evaluator import EvaluatorContext, EvaluatorResult
from ropt.events import Event
from ropt.plan import OptimizationPlanRunner
from ropt.results import FunctionResults

CONFIG: Dict[str, Any] = {
    "variables": {
        "initial_values": 2 * [0.0],
        "lower_bounds": [0.0, 0.0],
        "upper_bounds": [10.0, 10.0],
        "types": VariableType.INTEGER,
    },
    "optimizer": {
        "method": "differential_evolution",
        "options": {"integrality": [True, True], "seed": 4},
        "max_functions": 100,
        "parallel": True,
    },
    "linear_constraints": {
        "coefficients": [1.0, 1.0],
        "types": [ConstraintType.LE],
        "rhs_values": [10.0],
    },
}


def function(variables: NDArray[np.float64], _: EvaluatorContext) -> EvaluatorResult:
    """Evaluate the function.

    Args:
        variables: The variables to evaluate
        context:   Evaluator context

    Returns:
        Calculated objectives.
    """
    x = variables[:, 0]
    y = variables[:, 1]
    objectives = -np.array(np.minimum(3 * x, y), ndmin=2).T
    return EvaluatorResult(objectives=objectives)


def report(event: Event) -> None:
    """Report results of an evaluation.

    Args:
        event: event data
    """
    assert event.results is not None
    for item in event.results:
        if isinstance(item, FunctionResults) and item.functions is not None:
            print(f"result: {item.result_id}")
            print(f"  variables: {item.evaluations.variables}")
            print(f"  objective: {item.functions.weighted_objective}\n")


def run_optimization() -> None:
    """Run the optimization."""
    optimal_result = (
        OptimizationPlanRunner(CONFIG, function)
        .add_observer(EventType.FINISHED_EVALUATION, report)
        .run()
        .results
    )
    assert optimal_result is not None
    assert optimal_result.functions is not None
    assert np.all(optimal_result.evaluations.variables == [3, 7])
    print(f"BEST RESULT: {optimal_result.result_id}")
    print(f"  variables: {optimal_result.evaluations.variables}")
    print(f"  objective: {optimal_result.functions.weighted_objective}\n")


def main() -> None:
    """Main function."""
    run_optimization()


if __name__ == "__main__":
    main()



"""Script to run the rosenbrock optimization."""

# ruff: noqa: ERA001

import json
from pathlib import Path
from typing import Any, Dict

import numpy as np
from numpy.random import default_rng

from ropt.apps.script_optimizer import ScriptOptimizer

# Number of realizations:
REALIZATIONS = 3


ENOPT_CONFIG: Dict[str, Any] = {
    "variables": {
        "names": [("x", 1), ("y", 1)],
        "initial_values": [0.75, 1.25],
    },
    "optimizer": {
        "output_dir": "optimizer_output",
        "max_functions": 20,
        "speculative": True,
    },
    "objective_functions": {
        "names": ["rosenbrock"],
    },
    "realizations": {
        "names": [idx + 1 for idx in range(REALIZATIONS)],
        "weights": 1.0,
    },
    "gradient": {
        "perturbation_magnitudes": 1e-6,
    },
}

tasks = {
    "rosenbrock": """python ${work_dir}/../rosenbrock.py \
        --vars variables.json \
        --realization ${realization} \
        --coefficients ${work_dir}/../coefficients.json \
        --out rosenbrock""",
}

# Make data files for the problem:
rng = default_rng(seed=5)
a = rng.normal(loc=1.0, scale=0.1, size=REALIZATIONS)
b = rng.normal(loc=100.0, scale=10.0, size=REALIZATIONS)
coefficients = {
    realization: (a[idx], b[idx])
    for idx, realization in enumerate(ENOPT_CONFIG["realizations"]["names"])
}
with Path("coefficients.json").open("w", encoding="utf-8") as file_obj:
    json.dump(coefficients, file_obj, sort_keys=True, indent=4)


# Running locally:
provider = None

# # Example of running on a Slurm cluster:
# from parsl.providers import SlurmProvider
# provider = SlurmProvider(
#     max_blocks=16, exclusive=False, partition="defq", walltime="08:00:00"
# )


def main() -> None:
    """Run the example and check the result."""
    optimal_result = ScriptOptimizer(
        config={
            "work_dir": "work",
        },
        plan={
            "context": [
                {"id": "enopt_config", "init": "config", "with": ENOPT_CONFIG},
                {"id": "optimal", "init": "tracker"},
            ],
            "steps": [
                {
                    "run": "optimizer",
                    "with": {"config": "$enopt_config", "update": ["optimal"]},
                },
            ],
        },
        tasks=tasks,
    ).run(provider)["optimal"]
    if optimal_result is not None and optimal_result.functions is not None:
        print(f"BEST RESULT: {optimal_result.result_id}")
        print(f"  variables: {optimal_result.evaluations.variables}")
        print(f"  objective: {optimal_result.functions.weighted_objective}\n")
    assert optimal_result is not None
    assert optimal_result.functions is not None
    assert np.allclose(optimal_result.functions.weighted_objective, 0, atol=0.05)
    assert np.allclose(optimal_result.evaluations.variables, 1, atol=0.005)


if __name__ == "__main__":
    main()



"""This module implements the default optimizer step."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from pydantic import BaseModel, ConfigDict

from ropt.exceptions import PlanError
from ropt.plan import ContextUpdateDict
from ropt.plugins.plan.base import PlanStep

if TYPE_CHECKING:
    from ropt.config.plan import StepConfig
    from ropt.plan import Plan


class DefaultUpdateStepWith(BaseModel):
    """Parameters used by the default update step."""

    context: str
    value: Dict[str, Any]

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )


class DefaultUpdateStep(PlanStep):
    """The default update context step."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize a default update step.

        Args:
            config: The configuration of the step
            plan:   The plan that runs this step
        """
        super().__init__(config, plan)

        self._with = DefaultUpdateStepWith.model_validate(config.with_)

    def run(self) -> bool:
        """Run the update step.

        Returns:
            True if a user abort occurred, always `False`.
        """
        if not self.plan.has_context(self._with.context):
            msg = f"Env object `{self._with.context}` does not exist."
            raise PlanError(msg, step_name=self.step_config.name)
        self.plan.update_context(
            self._with.context,
            ContextUpdateDict(step_name=self.step_config.name, data=self._with.value),
        )
        return False



"""This module implements the default optimization plan plugin."""

from __future__ import annotations

from functools import singledispatchmethod
from typing import Dict, Final, Type, Union

from ropt.config.plan import ContextConfig, StepConfig  # noqa: TCH001
from ropt.plan import Plan  # noqa: TCH001
from ropt.plugins.plan.base import ContextObj, PlanStep  # noqa: TCH001

from ._config import DefaultConfigContext
from ._evaluator import DefaultEvaluatorStep
from ._optimizer import DefaultOptimizerStep
from ._repeat import DefaultRepeatStep
from ._reset import DefaultResetStep
from ._setvar import DefaultSetStep
from ._tracker import DefaultTrackerContext
from ._update import DefaultUpdateStep
from .base import PlanPlugin

_CONTEXT_OBJECTS: Final[Dict[str, Type[ContextObj]]] = {
    "config": DefaultConfigContext,
    "tracker": DefaultTrackerContext,
}

_STEP_OBJECTS: Final[Dict[str, Type[PlanStep]]] = {
    "evaluator": DefaultEvaluatorStep,
    "setvar": DefaultSetStep,
    "optimizer": DefaultOptimizerStep,
    "repeat": DefaultRepeatStep,
    "reset": DefaultResetStep,
    "update": DefaultUpdateStep,
}


class DefaultPlanPlugin(PlanPlugin):
    """Default plan plugin class."""

    @singledispatchmethod
    def create(  # type: ignore
        self,
        config: Union[ContextConfig, StepConfig],  # noqa: ARG002
        plan: Plan,  # noqa: ARG002
    ) -> Union[ContextObj, PlanStep]:
        """Initialize the plan plugin.

        See the [ropt.plugins.plan.base.PlanPlugin][] abstract base class.

        # noqa
        """
        msg = "Plan config type not implemented."
        raise NotImplementedError(msg)

    @create.register
    def _create_context(self, config: ContextConfig, plan: Plan) -> ContextObj:
        _, _, name = config.init.lower().rpartition("/")
        obj = _CONTEXT_OBJECTS.get(name)
        if obj is not None:
            return obj(config, plan)

        msg = f"Unknown context object type: {config.init}"
        raise TypeError(msg)

    @create.register
    def _create_step(self, config: StepConfig, plan: Plan) -> PlanStep:
        _, _, step_name = config.run.lower().rpartition("/")
        step_obj = _STEP_OBJECTS.get(step_name)
        if step_obj is not None:
            return step_obj(config, plan)

        msg = f"Unknown step type: {config.run}"
        raise TypeError(msg)

    def is_supported(self, method: str) -> bool:
        """Check if a method is supported.

        See the [ropt.plugins.base.Plugin][] abstract base class.

        # noqa
        """
        return (method.lower() in _CONTEXT_OBJECTS) or (method.lower() in _STEP_OBJECTS)



from __future__ import annotations

import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List

import numpy as np
import pytest
from pydantic import ValidationError

from ropt.config.plan import PlanConfig
from ropt.enums import EventType, OptimizerExitCode
from ropt.events import Event
from ropt.exceptions import PlanError
from ropt.plan import OptimizationPlanRunner, OptimizerContext, Plan
from ropt.results import FunctionResults, Results

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig

# ruff: noqa: SLF001


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "max_functions": 20,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_run_basic(enopt_config: Any, evaluator: Any) -> None:
    variables = OptimizationPlanRunner(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_invalid_context_ids() -> None:
    plan_config: Dict[str, Any] = {
        "context": [
            {
                "id": "1optimal",
                "init": "tracker",
            },
        ],
        "steps": [],
    }
    with pytest.raises(ValidationError, match=".*Invalid ID: 1optimal.*"):
        PlanConfig.model_validate(plan_config)


def test_duplicate_context_ids() -> None:
    plan_config: Dict[str, Any] = {
        "context": [
            {
                "id": "optimal",
                "init": "tracker",
            },
            {
                "id": "optimal",
                "init": "tracker",
            },
        ],
        "steps": [],
    }
    with pytest.raises(
        ValidationError, match=".*Duplicate Context ID\\(s\\): optimal.*"
    ):
        PlanConfig.model_validate(plan_config)


def test_parse_value(enopt_config: Any, evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "results",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["results"],
                },
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    assert plan.parse_value("${{ 1 }}") == 1
    assert plan.parse_value("${{ -1 }}") == -1
    assert not plan.parse_value("${{ not 1 }}")
    assert not plan.parse_value("${{ True and False }}")
    assert plan.parse_value("${{ True or False }}")
    assert plan.parse_value("${{ 1 + 1 }}") == 2
    assert plan.parse_value("${{ 2**3 }}") == 8
    assert plan.parse_value("${{ 3 % 2 }}") == 1
    assert plan.parse_value("${{ 3 // 2 }}") == 1
    assert plan.parse_value("${{ 2.5 + (2 + 3) / 2 }}") == 5
    assert plan.parse_value("${{ 1 < 2 }}")
    assert plan.parse_value("${{ 1 < 2 < 3 }}")
    assert not plan.parse_value("${{ 1 < 2 > 3 }}")

    assert plan.parse_value("$results") is None
    assert plan.parse_value("${{ [1, 2] }}") == [1, 2]
    assert plan.parse_value("${{ [$results, 2] }}") == [None, 2]

    assert plan.parse_value("a ${{ 1 }} b") == "a 1 b"
    assert plan.parse_value("a ${{ 1 + 1 }} b") == "a 2 b"
    assert plan.parse_value("a ${{ 1 + 1 }} b $results") == "a 2 b None"
    assert plan.parse_value("a ${{ 1 + 1 }} b $$results") == "a 2 b $results"

    with pytest.raises(
        PlanError,
        match=re.escape("Syntax error in expression: 1 + 1 ${{ x"),
    ):
        plan.parse_value("a $results ${{ 1 + 1 ${{ x }} }} b")

    with pytest.raises(
        PlanError, match=re.escape("Syntax error in expression: 1 + * 1")
    ):
        plan.parse_value("${{ 1 + * 1 }}")

    with pytest.raises(
        PlanError,
        match=re.escape("Unknown plan variable: `y`"),
    ):
        plan.parse_value("${{ $y + 1 }}")

    plan.run()

    assert isinstance(plan.parse_value("$results"), Results)


def test_setvar(evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "steps": [
            {
                "run": "setvar",
                "with": {
                    "var": "x",
                    "value": "${{ 1 }}",
                },
            },
            {
                "run": "setvar",
                "with": {
                    "var": "y",
                    "value": 1,
                },
            },
            {
                "run": "setvar",
                "with": {
                    "expr": "z = $y + 1",
                },
            },
            {
                "run": "setvar",
                "with": "u = 1",
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    assert plan["x"] == 1
    assert plan["y"] == 1
    assert plan["z"] == 2
    assert plan["u"] == 1


def test_invalid_setvar(evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "steps": [
            {
                "run": "setvar",
                "with": "1",
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    with pytest.raises(PlanError, match=re.escape("Invalid expression: 1")):
        Plan(parsed_config, context)

    plan_config = {
        "steps": [
            {
                "run": "setvar",
                "with": "2a = 1",
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    with pytest.raises(PlanError, match=re.escape("Invalid identifier: 2a")):
        Plan(parsed_config, context)


def test_invalid_identifier(evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "steps": [
            {"run": "setvar", "with": "x=1"},
            {
                "run": "setvar",
                "with": "y=x + 1",
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    with pytest.raises(PlanError, match=re.escape("Syntax error in expression: x + 1")):
        plan.run()


def test_conditional_run(enopt_config: EnOptConfig, evaluator: Any) -> None:
    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimal1",
                "init": "tracker",
            },
            {
                "id": "optimal2",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimal1"],
                },
                "if": "${{ 1 > 0 }}",
            },
            {"run": "setvar", "with": "x = 1"},
            {
                "run": "optimizer",
                "if": "$x < 0",
                "with": {
                    "config": "$config",
                    "update": ["optimal2"],
                },
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    result1 = plan["optimal1"]
    result2 = plan["optimal2"]
    assert result1 is not None
    assert np.allclose(result1.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    assert result2 is None


def test_set_initial_values(enopt_config: EnOptConfig, evaluator: Any) -> None:
    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimal1",
                "init": "tracker",
            },
            {
                "id": "optimal2",
                "init": "tracker",
            },
            {
                "id": "optimal3",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimal1"],
                },
            },
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimal2"],
                    "initial_values": "$optimal1",
                },
            },
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimal3"],
                    "initial_values": [0, 0, 0],
                },
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    result1 = plan["optimal1"]
    assert result1 is not None
    assert np.allclose(result1.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)

    result2 = plan["optimal2"]
    assert result2 is not None
    assert np.allclose(result2.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)

    result3 = plan["optimal2"]
    assert result3 is not None
    assert np.allclose(result3.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)

    assert not np.all(result1.evaluations.variables == result2.evaluations.variables)
    assert not np.all(result1.evaluations.variables == result3.evaluations.variables)


def test_reset_results(enopt_config: EnOptConfig, evaluator: Any) -> None:
    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimal",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimal"],
                },
            },
            {
                "run": "setvar",
                "with": "saved_results = $optimal",
            },
            {
                "run": "reset",
                "with": "optimal",
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["optimal"] is None
    saved_results = plan["saved_results"]
    assert saved_results is not None
    assert np.allclose(saved_results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_two_optimizers_alternating(enopt_config: Any, evaluator: Any) -> None:
    completed_functions = 0

    def _track_evaluations(event: Event) -> None:
        nonlocal completed_functions
        assert event.results is not None
        for item in event.results:
            if isinstance(item, FunctionResults):
                completed_functions += 1

    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    opt_config1 = {
        "speculative": True,
        "max_functions": 4,
    }
    opt_config2 = {
        "speculative": True,
        "max_functions": 3,
    }

    enopt_config1 = deepcopy(enopt_config)
    enopt_config1["variables"]["indices"] = [0, 2]
    enopt_config1["optimizer"] = opt_config1
    enopt_config2 = deepcopy(enopt_config)
    enopt_config2["variables"]["indices"] = [1]
    enopt_config2["optimizer"] = opt_config2

    plan_config = {
        "context": [
            {
                "id": "enopt_config1",
                "init": "config",
                "with": {
                    "config": enopt_config1,
                },
            },
            {
                "id": "enopt_config2",
                "init": "config",
                "with": enopt_config2,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
            {
                "id": "last",
                "init": "tracker",
                "with": {
                    "type": "last",
                },
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config1",
                    "update": ["last", "optimum"],
                },
            },
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config2",
                    "update": ["last", "optimum"],
                    "initial_values": "$last",
                },
            },
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config1",
                    "update": ["last", "optimum"],
                    "initial_values": "$last",
                },
            },
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config2",
                    "update": ["optimum"],
                    "initial_values": "$last",
                },
            },
        ],
    }

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()

    assert completed_functions == 14
    assert plan["optimum"] is not None
    assert np.allclose(
        plan["optimum"].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )


def test_optimization_sequential(enopt_config: Any, evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 2

    enopt_config2 = deepcopy(enopt_config)
    enopt_config2["optimizer"]["max_functions"] = 3

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "enopt_config2",
                "init": "config",
                "with": enopt_config2,
            },
            {
                "id": "last",
                "init": "tracker",
                "with": {
                    "type": "last",
                },
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config",
                    "update": ["last"],
                },
            },
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config2",
                    "initial_values": "$last",
                },
            },
        ],
    }

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()

    assert not np.allclose(
        completed[1].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )
    assert np.all(
        completed[2].evaluations.variables == completed[1].evaluations.variables
    )
    assert np.allclose(completed[-1].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_repeat_step(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config",
                    "update": ["optimum"],
                },
            },
        ],
    }
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()
    assert plan["optimum"] is not None
    variables = plan["optimum"].evaluations.variables.copy()
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)

    plan_config["steps"] = [
        {
            "run": "repeat",
            "with": {
                "iterations": 1,
                "steps": [
                    {
                        "run": "optimizer",
                        "with": {
                            "config": "$enopt_config",
                            "update": ["optimum"],
                        },
                    },
                ],
            },
        }
    ]
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()
    assert plan["optimum"] is not None

    assert np.all(variables == plan["optimum"].evaluations.variables)

    assert np.all(
        OptimizationPlanRunner(enopt_config, evaluator()).repeat(1).run().variables
        == variables
    )


def test_restart_initial(enopt_config: Any, evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 3

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
        ],
        "steps": [
            {
                "run": "repeat",
                "with": {
                    "iterations": 2,
                    "steps": [
                        {
                            "run": "optimizer",
                            "with": {
                                "config": "$enopt_config",
                            },
                        },
                    ],
                },
            }
        ],
    }

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()

    assert len(completed) == 6

    initial = np.array([enopt_config["variables"]["initial_values"]])
    assert np.all(completed[0].evaluations.variables == initial)
    assert np.all(completed[3].evaluations.variables == initial)

    completed = []
    OptimizationPlanRunner(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    ).repeat(2, restart_from="initial").run()
    assert np.all(completed[0].evaluations.variables == initial)
    assert np.all(completed[3].evaluations.variables == initial)


def test_restart_last(enopt_config: Any, evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 3

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "last",
                "init": "tracker",
                "with": {"type": "last"},
            },
        ],
        "steps": [
            {
                "run": "repeat",
                "with": {
                    "iterations": 2,
                    "steps": [
                        {
                            "run": "optimizer",
                            "with": {
                                "config": "$enopt_config",
                                "update": ["last"],
                                "initial_values": "$last",
                            },
                        },
                    ],
                },
            }
        ],
    }
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()

    assert np.all(
        completed[3].evaluations.variables == completed[2].evaluations.variables
    )

    completed = []
    OptimizationPlanRunner(enopt_config, evaluator()).repeat(
        2, restart_from="last"
    ).add_observer(EventType.FINISHED_EVALUATION, _track_evaluations).run()
    assert np.all(
        completed[3].evaluations.variables == completed[2].evaluations.variables
    )


def test_restart_optimum(enopt_config: Any, evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "repeat",
                "with": {
                    "iterations": 2,
                    "steps": [
                        {
                            "run": "optimizer",
                            "with": {
                                "config": "$enopt_config",
                                "update": ["optimum"],
                                "initial_values": "$optimum",
                            },
                        },
                    ],
                },
            }
        ],
    }
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()

    assert np.all(
        completed[2].evaluations.variables == completed[4].evaluations.variables
    )

    completed = []
    OptimizationPlanRunner(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    ).repeat(2, restart_from="optimal").run()
    assert np.all(
        completed[2].evaluations.variables == completed[4].evaluations.variables
    )


def test_restart_optimum_with_reset(
    enopt_config: Any, evaluator: Any, test_functions: Any
) -> None:
    completed: List[FunctionResults] = []
    max_functions = 5

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    # Make sure each restart has worse objectives, and that the last evaluation
    # is even worse, so each run has its own optimum that is worse than the
    # global and not at its last evaluation. We should end up with initial
    # values that are not from the global optimum, or from the most recent
    # evaluation:
    new_functions = (
        lambda variables, context: (
            test_functions[0](variables, context)
            + int((len(completed) + 1) / max_functions)
        ),
        lambda variables, context: (
            test_functions[1](variables, context)
            + int((len(completed) + 1) / max_functions)
        ),
    )

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = max_functions

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "repeat",
                "with": {
                    "iterations": 3,
                    "steps": [
                        {
                            "run": "setvar",
                            "with": "initial = $optimum",
                        },
                        {
                            "run": "reset",
                            "with": {
                                "context": "optimum",
                            },
                        },
                        {
                            "run": "optimizer",
                            "with": {
                                "config": "$enopt_config",
                                "update": ["optimum"],
                                "initial_values": "$initial",
                            },
                        },
                    ],
                },
            },
        ],
    }
    context = OptimizerContext(evaluator=evaluator(new_functions))
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()

    # The third evaluation is the optimum, and used to restart the second run:
    assert np.all(
        completed[max_functions].evaluations.variables
        == completed[2].evaluations.variables
    )
    # The 8th evaluation is the optimum of the second run, and used for the third:
    assert np.all(
        completed[2 * max_functions].evaluations.variables
        == completed[8].evaluations.variables
    )

    completed = []
    OptimizationPlanRunner(enopt_config, evaluator(new_functions)).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    ).repeat(3, restart_from="last_optimal").run()

    # The third evaluation is the optimum, and used to restart the second run:
    assert np.all(
        completed[max_functions].evaluations.variables
        == completed[2].evaluations.variables
    )
    # The 8th evaluation is the optimum of the second run, and used for the third:
    assert np.all(
        completed[2 * max_functions].evaluations.variables
        == completed[8].evaluations.variables
    )


def test_repeat_metadata(enopt_config: EnOptConfig, evaluator: Any) -> None:
    restarts: List[int] = []

    def _track_results(event: Event) -> None:
        assert event.results is not None
        metadata = event.results[0].metadata
        restart = metadata.get("restart", -1)
        assert metadata["foo"] == 1
        assert metadata["bar"] == "string"
        assert metadata["complex"] == f"string 2 {restart}"
        if not restarts or restart != restarts[-1]:
            restarts.append(restart)

    metadata = {
        "restart": "$counter",
        "foo": 1,
        "bar": "string",
        "complex": "string ${{ 1 + 1}} $counter",
    }

    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
        ],
        "steps": [
            {
                "run": "repeat",
                "with": {
                    "iterations": 2,
                    "counter_var": "counter",
                    "steps": [
                        {
                            "run": "optimizer",
                            "with": {
                                "config": "$config",
                                "metadata": metadata,
                            },
                        },
                    ],
                },
            }
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_results
    )
    plan.run()
    assert restarts == [0, 1]

    restarts = []
    OptimizationPlanRunner(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_results
    ).add_metadata(metadata).repeat(
        2, restart_from="last_optimal", counter_var="counter"
    ).run()
    assert restarts == [0, 1]


def test_update_enopt(enopt_config: Any, evaluator: Any) -> None:
    weights = enopt_config["objective_functions"]["weights"]
    enopt_config["objective_functions"]["weights"] = [1, 1]

    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimum"],
                },
            },
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["optimum"] is not None
    assert not np.allclose(
        plan["optimum"].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )

    plan_config["steps"] = [
        {
            "run": "update",
            "with": {
                "context": "config",
                "value": {"objective_functions": {"weights": weights}},
            },
        },
    ] + plan_config["steps"]
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    assert plan["optimum"] is not None
    assert np.allclose(
        plan["optimum"].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )


def test_evaluator_step(enopt_config: Any, evaluator: Any) -> None:
    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "evaluator",
                "with": {
                    "config": "$config",
                    "update": ["optimum"],
                },
            },
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["optimum"] is not None
    assert plan["optimum"].functions is not None
    assert np.allclose(plan["optimum"].functions.weighted_objective, 1.66)

    plan_config["steps"][0]["with"]["values"] = [0, 0, 0]
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["optimum"] is not None
    assert plan["optimum"].functions is not None
    assert np.allclose(plan["optimum"].functions.weighted_objective, 1.75)


def test_evaluator_step_multi(enopt_config: Any, evaluator: Any) -> None:
    completed: List[float] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item.functions.weighted_objective.item()
            for item in event.results
            if isinstance(item, FunctionResults) and item.functions is not None
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    plan_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "evaluator",
                "with": {
                    "config": "$config",
                    "update": ["optimum"],
                    "values": [[0, 0, 0.1], [0, 0, 0]],
                },
            },
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATOR_STEP, _track_evaluations
    )
    plan.run()

    assert len(completed) == 2
    assert np.allclose(completed, [1.66, 1.75])


def test_exit_code(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    is_called = False

    def _exit_code(
        event: Event,
    ) -> None:
        nonlocal is_called
        is_called = True
        assert isinstance(event, Event)
        assert event.exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED

    plan_config = {
        "context": [
            {
                "id": "enopt_config",
                "init": "config",
                "with": enopt_config,
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$enopt_config",
                    "exit_code_var": "exit_code",
                },
            },
        ],
    }
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_OPTIMIZER_STEP, _exit_code
    )
    plan.run()
    assert plan["exit_code"] == OptimizerExitCode.MAX_FUNCTIONS_REACHED
    assert is_called


def test_nested_plan(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    completed_functions = 0

    def _track_evaluations(event: Event) -> None:
        nonlocal completed_functions
        assert event.results is not None
        for item in event.results:
            if isinstance(item, FunctionResults):
                completed_functions += 1

    enopt_config["optimizer"]["tolerance"] = 1e-10
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4
    enopt_config["variables"]["indices"] = [0, 2]
    nested_config = deepcopy(enopt_config)
    nested_config["variables"]["indices"] = [1]
    enopt_config["optimizer"]["max_functions"] = 5

    inner_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": nested_config,
            },
            {
                "id": "nested_optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["nested_optimum"],
                    "initial_values": "$initial",
                },
            },
        ],
    }

    outer_config = {
        "context": [
            {
                "id": "config",
                "init": "config",
                "with": enopt_config,
            },
            {
                "id": "optimum",
                "init": "tracker",
            },
        ],
        "steps": [
            {
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimum"],
                    "nested_plan": {
                        "plan": inner_config,
                        "initial_var": "initial",
                        "results_var": "nested_optimum",
                    },
                },
            },
        ],
    }

    parsed_config = PlanConfig.model_validate(outer_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()
    results = plan["optimum"]

    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    assert completed_functions == 25


def test_nested_plan_metadata(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    def _track_evaluations(event: Event) -> None:
        assert event.results is not None
        for item in event.results:
            if isinstance(item, FunctionResults):
                assert item.metadata.get("outer") == 1
                if event.step_name == "inner":
                    assert item.metadata.get("inner") == "inner_meta_data"

    enopt_config["optimizer"]["tolerance"] = 1e-10
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4
    enopt_config["variables"]["indices"] = [0, 2]
    nested_config = deepcopy(enopt_config)
    nested_config["variables"]["indices"] = [1]
    enopt_config["optimizer"]["max_functions"] = 5

    inner_config = {
        "context": [
            {"id": "config", "init": "config", "with": nested_config},
            {"id": "nested_optimum", "init": "tracker"},
        ],
        "steps": [
            {
                "name": "inner",
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["nested_optimum"],
                    "initial_values": "$initial",
                    "metadata": {"inner": "inner_meta_data"},
                },
            },
        ],
    }

    outer_config = {
        "context": [
            {"id": "config", "init": "config", "with": enopt_config},
            {"id": "optimum", "init": "tracker"},
        ],
        "steps": [
            {"run": "setvar", "with": "x = 1"},
            {
                "name": "outer",
                "run": "optimizer",
                "with": {
                    "config": "$config",
                    "update": ["optimum"],
                    "nested_plan": {
                        "plan": inner_config,
                        "initial_var": "initial",
                        "results_var": "nested_optimum",
                    },
                    "metadata": {"outer": "$x"},
                },
            },
        ],
    }

    parsed_config = PlanConfig.model_validate(outer_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.optimizer_context.events.add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan.run()
    results = plan["optimum"]

    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)



"""This module defines the abstract base class for plugins."""

from __future__ import annotations

from abc import ABC, abstractmethod


class Plugin(ABC):
    """Abstract base class for plugins.

    `ropt`  plugins should derive from this base class, which specifies an
    `is_supported` method.
    """

    @abstractmethod
    def is_supported(self, method: str) -> bool:
        """Check wether a given method is supported.

        Args:
            method: The method name

        Returns:
            True if the method is supported.
        """



from __future__ import annotations

import json
from contextlib import suppress
from dataclasses import fields, is_dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Literal,
    Optional,
    Tuple,
    Type,
)

import xarray

from ropt.enums import ResultAxisName

if TYPE_CHECKING:
    from pathlib import Path

    from ropt.config.enopt import EnOptConfig

    from ._result_field import ResultField
    from ._results import Results, TypeResults


def _to_dataset(  # noqa: PLR0913
    config: EnOptConfig,
    result_field: ResultField,
    result_id: int,
    batch_id: Optional[int],
    metadata: Dict[str, Any],
    select: Optional[Iterable[str]],
) -> xarray.Dataset:
    if select is None:
        select = (field.name for field in fields(result_field))
    arrays = {field: _to_data_array(config, result_field, field) for field in select}
    arrays = {field: array for field, array in arrays.items() if array is not None}
    attrs: Dict[str, Any] = {}
    attrs["result_id"] = result_id
    if batch_id is not None:
        attrs["batch_id"] = batch_id
    if metadata:
        attrs["metadata"] = metadata
    return xarray.Dataset(arrays, attrs=attrs)


def _to_data_array(
    config: EnOptConfig, result_field: ResultField, field: str
) -> Optional[xarray.DataArray]:
    try:
        data = getattr(result_field, field)
    except AttributeError as exc:
        msg = f"Not a field name: {field}"
        raise ValueError(msg) from exc
    if data is None:
        return None
    axes = result_field.get_axis_names(field)
    indices = [_get_index(config, axis) for axis in axes]
    return xarray.DataArray(
        data,
        coords={
            axis.value: list(index)
            for axis, index in zip(axes, indices)
            if index is not None
        },
        dims=[axis.value for axis in axes],
    )


def _get_index(config: EnOptConfig, axis: ResultAxisName) -> Optional[Tuple[Any, ...]]:
    if axis == ResultAxisName.VARIABLE:
        return (
            None
            if config.variables.names is None
            else config.variables.get_formatted_names()
        )
    if axis == ResultAxisName.OBJECTIVE:
        return config.objective_functions.names
    if axis == ResultAxisName.NONLINEAR_CONSTRAINT:
        assert config.nonlinear_constraints is not None
        return config.nonlinear_constraints.names
    if axis == ResultAxisName.REALIZATION:
        return config.realizations.names
    return None


def _from_dataset(dataset: xarray.Dataset) -> Dict[str, Any]:
    return {str(name): data.to_numpy() for name, data in dataset.items()}


def _to_netcdf(results: Results, config: EnOptConfig, filename: Path) -> None:
    filename = filename.with_name(
        filename.name.format(result_id=results.result_id, batch_id=results.batch_id),
    )
    if filename.suffix != ".nc":
        filename = filename.with_suffix(".nc")
    mode: Literal["w", "a"] = "w"
    metadata: Dict[str, Any] = {
        "result_id": results.result_id,
        "metadata": json.dumps(results.metadata),
    }
    if results.batch_id is not None:
        metadata["batch_id"] = results.batch_id
    for result_field in fields(results):
        if is_dataclass(getattr(results, result_field.name)):
            dataset = results.to_dataset(config, result_field.name)
            dataset.to_netcdf(filename, mode=mode, group=result_field.name)
            mode = "a"
    if filename.exists():
        xarray.Dataset(attrs=metadata).to_netcdf(
            filename, mode="a", group="__metadata__"
        )


def _from_netcdf(filename: Path, result_type: Type[TypeResults]) -> Dict[str, Any]:
    if filename.suffix != ".nc":
        filename = filename.with_suffix(".nc")
    result: Dict[str, Any] = xarray.open_dataset(filename, group="__metadata__").attrs
    result["metadata"] = json.loads(result["metadata"])
    for result_field in fields(result_type):
        # Load all fields, skipping those already loaded from the metadata:
        if result_field.name not in result:
            with suppress(OSError):  # Not all fields have to be present
                result[result_field.name] = xarray.open_dataset(
                    filename, group=result_field.name
                )
    return result



"""This module defines the abstract base class for optimization steps."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field

from ropt.plan import ContextUpdate, ContextUpdateResults
from ropt.plugins.plan.base import ContextObj

from ._utils import _get_last_result, _update_optimal_result

if TYPE_CHECKING:
    from ropt.config.plan import ContextConfig
    from ropt.plan import Plan
    from ropt.results import FunctionResults


class DefaultTrackerWith(BaseModel):
    """Parameters for the track_results context object.

    The `type` parameter determines what result is tracked:
    - "optimal": Track the best result added
    - "last": Track the last result added

    Attributes:
        type:                 The type of result to store
        constraint_tolerance: Optional constraint tolerance
    """

    type_: Literal["optimal", "last"] = Field(default="optimal", alias="type")
    constraint_tolerance: Optional[float] = 1e-10

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )


class DefaultTrackerContext(ContextObj):
    """The default results context object."""

    def __init__(self, config: ContextConfig, plan: Plan) -> None:
        """Initialize a default tracker context object.

        Args:
            config: The configuration of the step
            plan:   The plan that runs this step
        """
        super().__init__(config, plan)
        self._with = (
            DefaultTrackerWith()
            if config.with_ is None
            else DefaultTrackerWith.model_validate(config.with_)
        )
        self.set_variable(None)

    def update(self, value: ContextUpdate) -> None:
        """Update the tracker object.

        Updates the stored results. If the `constraint_tolerance` value set in
        the object configuration is `None`, the object value will be updated
        according to the `type`. If `constraint_tolerance` is not `None`, it
        will only be updated if the results do not violate any constraints.

        Args:
            value: The value to set.
        """
        if isinstance(value, ContextUpdateResults):
            results: Optional[FunctionResults] = None
            if self._with.type_ == "optimal":
                results = _update_optimal_result(
                    self.get_variable(), value.results, self._with.constraint_tolerance
                )
            if self._with.type_ == "last":
                results = _get_last_result(
                    value.results, self._with.constraint_tolerance
                )
            if results is not None:
                results = deepcopy(results)
                self.set_variable(results)

    def reset(self) -> None:
        """Clear the stored values."""
        self.set_variable(None)



"""This module defines default configuration context object."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, Union

from pydantic import BaseModel, ConfigDict

from ropt.config.enopt import EnOptConfig
from ropt.plan import ContextUpdate, ContextUpdateDict
from ropt.plugins.plan.base import ContextObj

if TYPE_CHECKING:
    from ropt.config.plan import ContextConfig
    from ropt.plan import Plan


class DefaultConfigWith(BaseModel):
    """Parameters for the `config` context object.

    Attributes:
        config: The configuration
    """

    config: Union[EnOptConfig, Dict[str, Any]]

    model_config = ConfigDict(
        extra="forbid",
        validate_default=True,
    )


class DefaultConfigContext(ContextObj):
    """The default `config` context object.

    The `config` object is used store and modify configuration objects for
    optimizer and evaluation steps. When initializing the object, the `with`
    field is parsed as an [`EnOptConfig`][ropt.config.enopt.EnOptConfig] object.
    Optimization and evaluation steps can refer to an optimization plan variable
    of the same name as the `config` ID to retrieve the parsed configuration
    object.

    The `config` object can be updated with a dictionary of values to construct
    a new `EnOptConfig` object with the updated values. The original value is
    kept with the `config` object and can be recovered by resetting the `config`
    object.
    """

    def __init__(self, config: ContextConfig, plan: Plan) -> None:
        super().__init__(config, plan)
        if "config" in config.with_:
            with_ = DefaultConfigWith.model_validate(config.with_)
            enopt_config = EnOptConfig.model_validate(with_.config)
        else:
            enopt_config = EnOptConfig.model_validate(config.with_)

        self._backup = enopt_config
        self.set_variable(enopt_config)

    def update(self, updates: ContextUpdate) -> None:
        if isinstance(updates, ContextUpdateDict):
            enopt_config = self.get_variable()
            assert enopt_config.original_inputs is not None
            enopt_config = EnOptConfig.model_validate(
                _update_dict(enopt_config.original_inputs, updates.data)
            )
            self.set_variable(enopt_config)

    def reset(self) -> None:
        """Clear the stored values."""
        self.set_variable(self._backup)


def _update_dict(mapping: Dict[str, Any], values: Dict[str, Any]) -> Dict[str, Any]:
    mapping = mapping.copy()
    for key, value in values.items():
        if (
            key in mapping
            and isinstance(mapping[key], dict)
            and isinstance(value, dict)
        ):
            mapping[key] = _update_dict(mapping[key], value)
        else:
            mapping[key] = deepcopy(value)
    return mapping



"""This module defines the base classes for optimization plan objects.

Optimization plan objects can be added via the plugin mechanism to implement
additional functionality. This is done by creating a plugin class that derives
from the [`PlanPlugin`][ropt.plugins.plan.base.PlanPlugin] class. It
needs to define a [`create`][ropt.plugins.plan.base.PlanPlugin].create
method that generates the plan objects.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union

from ropt.plugins.base import Plugin

if TYPE_CHECKING:
    from ropt.config.plan import ContextConfig, StepConfig
    from ropt.plan import ContextUpdate, OptimizerContext, Plan


class ContextObj(ABC):
    """Base class for context objects."""

    def __init__(self, config: ContextConfig, plan: Plan) -> None:
        """Initialize the context object.

        The `config` and `plan` arguments are accessible as `context_config`
        and `plan` properties.

        Args:
            config: The configuration of the context object
            plan:   The parent plan that contains the object
        """
        self._context_config = config
        self._plan = plan

    @abstractmethod
    def update(self, value: ContextUpdate) -> None:
        """Update the value of the object.

        Args:
            value: The value used for the update.
        """

    def reset(self) -> None:  # noqa: B027
        """Resets the object to its initial state."""

    @property
    def context_config(self) -> ContextConfig:
        """Return the context object configuration.

        Returns:
            The configuration object.
        """
        return self._context_config

    @property
    def plan(self) -> Plan:
        """Return the plan object.

        Returns:
            The plan object.
        """
        return self._plan

    def get_variable(self) -> Any:  # noqa: ANN401
        """Get a variable with the name equal to the context object ID.

        Returns:
            The value of the variable.
        """
        return self._plan[self._context_config.id]

    def set_variable(self, value: Any) -> None:  # noqa: ANN401
        """Set a variable with the name equal to the context object ID.

        Args:
            value: The value
        """
        self._plan[self._context_config.id] = value


class PlanStep(ABC):
    """Base class for plan steps."""

    def __init__(self, config: StepConfig, plan: Plan) -> None:
        """Initialize the plan object.

        Args:
            config: The configuration of the plan object
            plan:   The parent plan
        """
        self._step_config = config
        self._plan = plan

    @property
    def step_config(self) -> StepConfig:
        """Return the step object config.

        Returns:
            The configuration object.
        """
        return self._step_config

    @property
    def plan(self) -> Plan:
        """Return the plan object.

        Returns:
            The plan object.
        """
        return self._plan

    @abstractmethod
    def run(self) -> bool:
        """Run the step object.

        Returns:
            `True` if a user abort occurred.
        """


class PlanPlugin(Plugin):
    """The abstract base class for plan plugins."""

    @abstractmethod
    def create(
        self, config: Union[ContextConfig, StepConfig], context: OptimizerContext
    ) -> Union[ContextObj, PlanStep]:
        """Create the plan object.

        Args:
            config:  The configuration of the plan object
            context: The context in which the plan operates
        """



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from importlib.util import find_spec
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Final, Iterable, Optional, TypeVar, Union

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig
    from ropt.enums import ResultAxisName


_HAVE_PANDAS: Final = find_spec("pandas") is not None
_HAVE_XARRAY: Final = find_spec("xarray") is not None
_HAVE_NETCDF: Final = find_spec("netCDF4") is not None

if TYPE_CHECKING and _HAVE_PANDAS:
    import pandas as pd  # noqa: TCH002
if _HAVE_PANDAS:
    from ._pandas import _to_dataframe
if TYPE_CHECKING and _HAVE_XARRAY:
    import xarray  # noqa: TCH002
if _HAVE_XARRAY:
    from ._xarray import _to_dataset
if _HAVE_XARRAY and _HAVE_NETCDF:
    from ._xarray import _to_netcdf

TypeResults = TypeVar("TypeResults", bound="Results")


@dataclass
class Results(ABC):
    """The `Results` class serves as an abstract base class for storing results.

    This class contains the following generic information:

    1. The result ID, a unique integer value that is incremented as the
       results become available.
    2. An optional batch ID, which may be generated by the function evaluator.
       The interpretation of this ID depends on the evaluator code. It is
       intended to be used as a unique identifier for the group of function
       evaluations passed to the evaluator by the optimization code.
    3. A dictionary of metadata to be added by optimization steps. This contains
       generic information, the nature of which depends on the steps producing
       them. They are expected to be primitive values not interpreted by the
       optimization code but can be exported and reported.

    The `Results` class is an abstract base class that is not intended to be
    instantiated by itself. Most data of interest will be stored in additional
    fields in one of the derived classes:
    [`FunctionResults`][ropt.results.FunctionResults] or
    [`GradientResults`][ropt.results.GradientResults]. In addition to the
    attributes containing the data, a few methods are provided to export the
    data. These functions will only be useful when used with one of the derived
    classes:

    1. The [`to_dataframe`][ropt.results.Results.to_dataframe] method can be
       used to export the contents, or a sub-set, of a single field to a
       [`pandas`](https://pandas.pydata.org/) data frame.
    2. The [`to_dataset`][ropt.results.Results.to_dataset] method can be used to
       export the contents, or a sub-set, of a single field to an
       [`xarray`](https://xarray.dev/) data set.
    2. The [`to_netcdf`][ropt.results.Results.to_netcdf] method can be used to
       export the entire results object to a
       [`netCDF`](https://www.unidata.ucar.edu/software/netcdf/) file.

    Attributes:
        result_id: The ID of the function/gradient evaluation.
        batch_id:  The ID of the evaluation batch that contains the result.
        metadata:  The metadata.
    """

    result_id: int
    batch_id: Optional[int]
    metadata: Dict[str, Any]

    def to_dataframe(
        self,
        config: EnOptConfig,
        field_name: str,
        select: Optional[Iterable[str]] = None,
        unstack: Optional[Iterable[ResultAxisName]] = None,
    ) -> pd.DataFrame:
        """Export a field to a pandas dataframe.

        The function exports the values of a single field to a pandas data
        frame. The field to export is selected by the `field_name` argument. In
        general such a field is another object with multiple sub-fields. By
        default, these are all exported as columns in the pandas data frame, but
        a sub-set can be selected using the `select` argument.

        Any of the sub-fields in the field that is exported may be a
        multi-dimensional array, which is exported in a stacked manner. Using
        the axis types found in the metadata, the exporter will construct a
        multi-index labeled with the corresponding names found in the optimizer
        configuration (if available, otherwise numerical indices are used). Such
        multi-indices can optionally be unstacked into multiple columns by
        providing the axis types to unstack via the `unstack` argument.

        Info: The data frame index
            As noted above, the index of the resulting data frame may be a
            multi-index constructed from axis indices or labels. In addition,
            the `result_id` field, and the `batch_id` (if not None), are also
            prepended to the index of the resulting frame. This has the
            beneficial effect that the data frames exported from multiple
            results can be concatenated and identified in the frame by result ID
            and/or batch ID.

        Args:
            config:     The ensemble optimizer configuration object
            field_name: The field to export
            select:     Select the sub-fields to export, by default all fields
            unstack:    Select axes to unstack, by default none

        Raises:
            NotImplementedError: If the pandas module is not installed
            ValueError:          If the field name is incorrect

        Returns:
            A pandas data frame containing the results.

        Warning:
            This function is only available if `pandas` is installed.
        """
        if not _HAVE_PANDAS:
            msg = "The pandas module must be installed to use to_dataframe"
            raise NotImplementedError(msg)

        result_field = getattr(self, field_name, None)
        if result_field is None:
            msg = f"Invalid result field: {field_name}"
            raise ValueError(msg)

        return _to_dataframe(
            config,
            result_field,
            self.result_id,
            self.batch_id,
            select,
            unstack,
        )

    def to_dataset(
        self,
        config: EnOptConfig,
        field_name: str,
        select: Optional[Iterable[str]] = None,
        *,
        add_metadata: bool = False,
    ) -> xarray.Dataset:
        """Export a field to an xarray dataset.

        The function exports the values of a single field to an xarray dataset.
        The field to export is selected by the `field_name` argument. In
        general, such a field is another object with multiple sub-fields. By
        default, these are all exported as data arrays in the xarray dataset,
        but a subset can be selected using the `select` argument.

        Any of the sub-fields in the field that is exported may be a
        multi-dimensional array, which is exported directly as an xarray data
        array. Using the axis types found in the metadata, the exporter will add
        coordinate labels constructed from the corresponding names found in the
        optimizer configuration (if available).

        The result ID and the batch ID (if not None) are added to the attributes
        of the dataset, using the `result_id` and `batch_id` keys. If metadata
        is present in the results, and the `add_metadata` flag is set, it is
        also added under the `metadata` key.

        Args:
            config:       The ensemble optimizer configuration object
            field_name:   The field to export
            select:       Select the fields to export; by default, all fields
            add_metadata: If true, add the metadata as a field in the dataset attrs

        Raises:
            NotImplementedError: If the `xarray` module is not installed.
            ValueError:          If the field name is incorrect.

        Returns:
            An xarray dataset containing the results.

        Warning:
            This function is only available if `xarray` is installed.
        """
        if not _HAVE_XARRAY:
            msg = "The xarray module must be installed to use to_dataset"
            raise NotImplementedError(msg)

        result_field = getattr(self, field_name, None)
        if result_field is None:
            msg = f"Invalid result field: {field_name}"
            raise ValueError(msg)

        return _to_dataset(
            config,
            result_field,
            self.result_id,
            self.batch_id,
            self.metadata if add_metadata else {},
            select,
        )

    def to_netcdf(self, config: EnOptConfig, filename: Union[str, Path]) -> None:
        """Write the results to a netCDF4 file.

        The fields of the result are converted to xarray datasets and each
        stored as a group in a netCDF4 file, using the name of the field as the
        group name. In addition, a special `__metadata__` group is added that
        contains the result ID, the batch ID (if not None), and a JSON dump of
        the metadata.

        By default, a file is written with a `.nc` extension. The filename may
        contain a replacement field for a variable named `result_id` or
        `batch_id`, which will be replaced by the corresponding field using the
        `format` string method. For example:
        `filename="results-{result_id:04d}"` will result in a file named
        `results-0001.nc` if the result ID equals 1.

        Info: Reading results from a netCDF4 file
            The results may be read back using the `from_netcdf` class method of
            either [`FunctionResults`][ropt.results.FunctionResults] or
            [`GradientResults`][ropt.results.GradientResults]. Currently, no
            information is written about whether the results were function or
            gradient results. Therefore, to read back the result with the
            correct class, its type must be known.

        Args:
            config:   The configuration used to run the optimization.
            filename: The name of the file to write.

        Raises:
            NotImplementedError:
                If the `xarray` or the `netCDF4` module is not installed.

        Warning:
            Use of this method requires that the `xarray` and `netCDF4` modules
            are installed.
        """
        if not _HAVE_XARRAY or not _HAVE_NETCDF:
            msg = (
                "The xarray and netCDF4 modules must be installed "
                "to use Results.to_netcdf"
            )
            raise NotImplementedError(msg)
        _to_netcdf(self, config, Path(filename))



"""Utilities used by the script-based optimizer."""

from collections import defaultdict
from itertools import groupby
from typing import Any, DefaultDict, List, Sequence, Tuple

import numpy as np
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig


def _make_dict(
    variables: NDArray[np.float64], names: Sequence[Tuple[str, ...]]
) -> DefaultDict[str, Any]:
    def recursive_dict() -> DefaultDict[str, Any]:
        return defaultdict(recursive_dict)

    var_dict = recursive_dict()
    for idx, var_name in enumerate(names):
        tmp = var_dict
        for name in var_name[:-1]:
            tmp = tmp[str(name)]
        tmp[str(var_name[-1])] = variables[idx]

    return var_dict


def _get_function_files(config: EnOptConfig) -> Tuple[Tuple[str, ...], Tuple[str, ...]]:
    assert config.objective_functions.names is not None
    objective_names = tuple(str(name) for name in config.objective_functions.names)
    if config.nonlinear_constraints is not None:
        assert config.nonlinear_constraints.names is not None
        constraint_names = tuple(
            str(name) for name in config.nonlinear_constraints.names
        )
    else:
        constraint_names = ()
    return objective_names, constraint_names


def _format_list(values: List[int]) -> str:
    grouped = (
        tuple(y for _, y in x)
        for _, x in groupby(enumerate(sorted(values)), lambda x: x[0] - x[1])
    )
    return ", ".join(
        "-".join([str(sub_group[0]), str(sub_group[-1])])
        if len(sub_group) > 1
        else str(sub_group[0])
        for sub_group in grouped
    )



"""The optimization event class."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

from ropt.enums import EventType, OptimizerExitCode

if TYPE_CHECKING:
    from ropt.config.enopt import EnOptConfig
    from ropt.results import Results


@dataclass
class Event:
    """The `Event` class stores optimization event data.

    While running an optimization plan, callbacks can be connected to react to
    events triggered during execution. These callbacks accept a single `Event`
    object containing information about the event.

    The actual data contained in the object depends on the nature of the event.
    Refer to the documentation of the [`EventType`][ropt.enums.EventType]
    enumeration for more details.

    Attributes:
        event_type: The type of the event
        config:     The current configuration object of the executing plan
        results:    Optional results passed with the event
        exit_code:  An optional exit code
    """

    event_type: EventType
    config: EnOptConfig
    results: Optional[Tuple[Results, ...]] = None
    exit_code: Optional[OptimizerExitCode] = None
    step_name: Optional[str] = None


class EventBroker:
    """A class for handling events."""

    def __init__(self) -> None:
        """Initialize the event broker."""
        self._subscribers: Dict[EventType, List[Callable[[Event], None]]] = {
            event: [] for event in EventType
        }

    def add_observer(
        self,
        event: EventType,
        callback: Callable[[Event], None],
    ) -> None:
        """Add an observer function.

        Args:
            event:    The type of events to react to
            callback: The function to call if the event is received
        """
        self._subscribers[event].append(callback)

    def emit(self, event_type: EventType, /, **kwargs: Any) -> None:  # noqa: ANN401
        """Emit an event.

        The keyword arguments are used to construct an
        [`Event`][ropt.events.Event] object of the type given by `event_type`.
        All stored callbacks that react to this event type are then called with
        that event object as their argument.

        Args:
            event_type: The type of event to emit
            kwargs:     Keyword arguments used to create an optimization event
        """
        event = Event(event_type=event_type, **kwargs)
        for callback in self._subscribers[event_type]:
            callback(event)



from pathlib import Path
from typing import Any, Dict

import numpy as np
import pytest

from ropt.config.enopt import EnOptConfig
from ropt.results import (
    FunctionEvaluations,
    FunctionResults,
    Functions,
    GradientEvaluations,
    GradientResults,
    Gradients,
    Realizations,
)

pytest.importorskip("xarray")
netcdf = pytest.importorskip("netCDF4")


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "variables": {
            "names": ["x", "y"],
            "initial_values": [0.0, 0.0],
        },
        "objective_functions": {
            "names": ["f1", "f2"],
            "weights": [0.75, 0.25],
        },
        "realizations": {
            "names": ["r1", "r2", "r3"],
            "weights": [1.0] * 3,
        },
        "gradient": {
            "number_of_perturbations": 5,
        },
    }


@pytest.fixture(name="function_result")
def function_result_fixture(enopt_config: Any) -> FunctionResults:
    config = EnOptConfig.model_validate(enopt_config)
    evaluations = FunctionEvaluations.create(
        config=config,
        variables=np.array([1.0, 2.0]),
        objectives=np.arange(6, dtype=np.float64).reshape((3, 2)),
    )
    realizations = Realizations(
        objective_weights=np.arange(6, dtype=np.float64).reshape((2, 3)),
        failed_realizations=np.zeros(3, dtype=np.bool_),
    )
    functions = Functions.create(
        config=config,
        weighted_objective=np.array(1.0),
        objectives=np.array([1.0, 2.0]),
    )
    return FunctionResults(
        result_id=0,
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=realizations,
        functions=functions,
    )


@pytest.fixture(name="gradient_result")
def gradient_result_fixture() -> GradientResults:
    evaluations = GradientEvaluations(
        variables=np.array([1.0, 2.0]),
        perturbed_variables=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
        perturbed_objectives=np.arange(30, dtype=np.float64).reshape((3, 5, 2)),
    )
    gradients = Gradients(
        weighted_objective=np.array([1.0, 2.0]),
        objectives=np.arange(4, dtype=np.float64).reshape((2, 2)),
    )
    return GradientResults(
        result_id=0,
        batch_id=1,
        metadata={},
        evaluations=evaluations,
        realizations=Realizations(
            failed_realizations=np.zeros(3, dtype=np.bool_),
        ),
        gradients=gradients,
    )


@pytest.mark.parametrize(
    ("name", "final_name"),
    [
        ("test", "test.nc"),
        ("test.nc", "test.nc"),
        ("test{batch_id:03d}", "test001.nc"),
        ("test{batch_id:03d}.nc", "test001.nc"),
    ],
)
def test_functions_to_netcdf(
    enopt_config: Any,
    function_result: FunctionResults,
    tmp_path: Path,
    name: str,
    final_name: str,
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    function_result.to_netcdf(config, tmp_path / name)
    assert (tmp_path / final_name).exists()
    file = netcdf.Dataset((tmp_path / final_name), mode="r")
    assert set(file.groups) == {
        "evaluations",
        "realizations",
        "functions",
        "__metadata__",
    }


def test_functions_from_netcdf(
    enopt_config: Any, function_result: FunctionResults, tmp_path: Path
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    function_result.to_netcdf(config, tmp_path / "test.nc")
    assert (tmp_path / "test.nc").exists()
    loaded = FunctionResults.from_netcdf(tmp_path / "test.nc")
    assert loaded.batch_id == function_result.batch_id
    for item in ("evaluations", "realizations", "functions"):
        assert loaded.to_dataframe(config, item).equals(
            function_result.to_dataframe(config, item)
        )


@pytest.mark.parametrize(
    ("name", "final_name"),
    [
        ("test", "test.nc"),
        ("test.nc", "test.nc"),
        ("test{batch_id:03d}", "test001.nc"),
        ("test{batch_id:03d}.nc", "test001.nc"),
    ],
)
def test_gradients_to_netcdf(
    enopt_config: Any,
    gradient_result: GradientResults,
    tmp_path: Path,
    name: str,
    final_name: str,
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    gradient_result.to_netcdf(config, tmp_path / name)
    assert (tmp_path / final_name).exists()
    file = netcdf.Dataset((tmp_path / final_name), mode="r")
    assert set(file.groups) == {
        "evaluations",
        "realizations",
        "gradients",
        "__metadata__",
    }


def test_gradients_from_netcdf(
    enopt_config: Any, gradient_result: GradientResults, tmp_path: Path
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    gradient_result.to_netcdf(config, tmp_path / "test.nc")
    assert (tmp_path / "test.nc").exists()
    loaded = GradientResults.from_netcdf(tmp_path / "test.nc")
    assert loaded.batch_id == gradient_result.batch_id
    for item in ("evaluations", "gradients"):
        assert loaded.to_dataframe(config, item).equals(
            gradient_result.to_dataframe(config, item)
        )


def test_netcdf_metadata(
    enopt_config: Any, function_result: FunctionResults, tmp_path: Path
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    function_result.metadata = {"foo": 1}
    function_result.to_netcdf(config, tmp_path / "test.nc")
    assert (tmp_path / "test.nc").exists()
    loaded = FunctionResults.from_netcdf(tmp_path / "test.nc")
    assert loaded.metadata == {"foo": 1}
